/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)37       1.82  src/avs/fs/mmfs/ts/kernext/gpl-linux/cxiIOBuffer.c, mmfs, avs_rgpfs32, rgpfs32s002a 10/15/07 10:42:43 */
/*
 * Linux implementation of I/O buffers
 *
 * Contents:
 *   struct cxiBufHeadChunk_t
 *   struct cxiIODoneQueue_t
 *
 *   IOBufferModuleInit
 *   IOBufferModuleTerm
 *   cxiPinMM
 *   cxiUnpinMM
 *
 *   cxiAllocPageList
 *   cxiMapAndRecordPages
 *   cxiReleaseAndForgetPages
 *   cxiGetPagePtrs
 *   cxiDeallocPageList
 *
 *   cxiAttachIOBuffer
 *   cxiDetachIOBuffer
 *   cxiUXfer
 *   cxiKXfer
 *   cxiKZero
 *   cxiMapDiscontiguousRW
 *   cxiUnmapDiscontiguousRW
 *   cxiMapContiguousRO
 *   cxiUnmapContiguousRO
 *
 *   chunkListDone
 *
 *  2.4 kernel:
 *   bufferHeadDone
 *   cxiStartIO
 *   GetDiskInfoX
 *
 *  2.6 kernel:
 *   bioDone
 *   cxiStartIO
 *   GetDiskInfoX
 *
 *   cxiWaitIO
 *
 *   cxiInitIODoneQueue
 *   cxiAllocIOWaitQueue
 *   cxiFreeIOWaitQueue
 *   cxiAllocIOCompletionQueueArray
 *   cxiFreeIOCompletionQueueArray
 */

#include <Shark-gpl.h>

#include <linux/module.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/smp_lock.h>
#if LINUX_KERNEL_VERSION >= 2050000
#include <linux/bio.h>
#else
#include <linux/iobuf.h>
#endif

#include <Logger-gpl.h>
#include <Trace.h>
#include <cxiSystem.h>
#include <linux2gpfs.h>
#include <verdep.h>
#include <cxiIOBuffer.h>
#include <cxiAtomic.h>
#include <cxiTypes.h>
#include <linux/mman.h>

//#define PAGE_ALIGN_DEBUG

#ifdef CONFIG_BGL
/* BG/L version of Linux doesn't define get_user_pages, so define it here */
#define get_user_pages(tsk, mm, start, len, write, force, pages, vmas) \
        __get_user_pages(tsk, mm, start, len, write, force, pages, vmas, 0)
#endif


/* Count of number of delays in busy wait loop in cxiWaitIO */
atomic_t cxiWaitIONDelays;

/* Static pointer to slab allocator for cxiBufHeadChunk_t's */
KMEM_CACHE_T* BhcCacheP = NULL;


/* Group of Linux buffer_heads or bios allocated together for a
   multi-page I/O.  A chunk of buffer heads is just less than half a
   page, while a chunk of bios is smaller. */
#if LINUX_KERNEL_VERSION >= 2050000
#define BIOS_PER_CHUNK \
  ((PAGE_SIZE/8 -      \
   (sizeof(struct list_head)+sizeof(atomic_t)+4*sizeof(void*)+sizeof(int))) / \
  sizeof(void*))
#else
#define BUFFER_HEADS_PER_CHUNK \
  ((PAGE_SIZE/2-               \
   (sizeof(struct list_head)+sizeof(atomic_t)+4*sizeof(void*)+sizeof(int))) / \
  sizeof(struct buffer_head))
#endif

/* A chunk of bio pointers or buffer_heads.  Allocated from the BhcCacheP
   slab cache. */
struct cxiBufHeadChunk_t
{
  /* Fields used only in the first chunk */

  /* Next and previous chunk lists on a cxiIODoneQueue.  The list is
     circular. */
  struct list_head bhcDoneLink;

  /* Pointer to I/O completion queue on which to enqueue this chunk list
     once all of its I/O has completed */
  struct cxiIODoneQueue_t*  bhcCompletionQueueP;

  /* Callback routine and data pointer for when I/O completes.  The
     ...Int... routine is called at interrupt level as soon as the I/O
     completes. */
  cxiIODoneIntCallback_t bhcIntCallbackRtn;
  void* bhcCallbackDataP;

  /* Total number of iodone calls remaining before all I/O is complete
     and this list of chunks can be moved to a cxiIODoneQueue.  Always
     updated with atomic operations, since this field is accessed
     asynchronously from interrupt level.*/
  atomic_t nIODoneLeft;

  /* Fields used in every chunk */

  /* Next chunk of bio pointers or buffer_heads used for a single I/O.
     The list is circular. */
  struct cxiBufHeadChunk_t*  bhcNextP;

  /* Number of buffer_heads or bios used in this chunk */
  int nSlotsUsed;

  /* Depending on kernel version, either an array of pointers to bios or
     an array of buffer_heads */
#if LINUX_KERNEL_VERSION >= 2050000
  struct bio *biop[BIOS_PER_CHUNK];
#else
  struct buffer_head bh[BUFFER_HEADS_PER_CHUNK];
#endif
};


/* Queue of lists of cxiBufHeadChunk_ts whose I/Os have completed. */
struct cxiIODoneQueue_t
{
  /* Wait queue used to wait for some number of I/Os to complete */
  wait_queue_head_t ioqWaitQ;

  /* Number of completed I/Os on this list.  Updated from iodone callbacks
     and from process context within cxiWaitIO. */
  atomic_t ioqNQueued;

  /* Number of completed I/Os before awakening the thread waiting in
     cxiWaitIO.  Only modified within cxiWaitIO. */
  atomic_t ioqNDesired;
#define MAX_IO_CHUNK_LISTS 65536

  /* First and last chunk lists of completed I/Os.  The list
     is circular, using the bhcDoneLink list links.  Protected by
     spinlock_irq_save on ioqWaitQ.lock. */
  struct list_head ioqListHead;
};


/* Initialization routine - called when module is loaded */
void 
IOBufferModuleInit()
{
  int rc;

  ENTER(0);
  TRACE0(TRACE_KSVFS, 1, TRCID_KIBD_INIT,
         "IOBufferModuleInit called\n");

  /* Create a slab allocator for cxiBufHeadChunk_t objects */
  BhcCacheP = kmem_cache_create("gpfsBufChunk",
                                 sizeof(struct cxiBufHeadChunk_t),
                                 0 /* offset */,
                                 0 /* flags */,
                                 NULL /* ctor */
#if LINUX_KERNEL_VERSION < 2062300
                               , NULL /* dtor */
#endif
                                 );
  if (BhcCacheP == NULL)
    cxiPanic("Cannot create cxiBufHeadChunk_t cache\n");

#if LINUX_KERNEL_VERSION >= 2060000
  if (gpfs_init_inodecache()!=0)
    cxiPanic("Cannot create gpfsInodeCache cache\n");
#endif

  atomic_set(&cxiWaitIONDelays, 0);
  EXIT(0);
}

/* Termination routine - called just before module is unloaded */
void 
IOBufferModuleTerm()
{
  int rc;

  ENTER(0);
  TRACE0(TRACE_KSVFS, 1, TRCID_KIBD_TERM,
         "IOBufferModuleTerm called\n");

  /* Destroy slab allocator for cxiBufHeadChunk_t objects */
#if LINUX_KERNEL_VERSION < 2061800
  rc = kmem_cache_destroy(BhcCacheP);
  if (rc)
    TRACE1(TRACE_KSVFS, 1, TRCID_BHC_CACHE_DESTROY,
           "IOBufferModuleTerm: ERROR! BufHeadChunk cache destroy rc %d", rc);
#else
  kmem_cache_destroy(BhcCacheP);
#endif

#if LINUX_KERNEL_VERSION >= 2060000
  gpfs_destroy_inodecache();
#endif
  EXIT(0);
}


#ifdef MMAP_DIO
/* Create a cross-memory descriptor object for a page in user address
   space that is already pinned.  The page will be mapped into kernel
   address space.  This is used by mmap routines that want to do
   direct I/O from user page to disk.  The cxiXmem_t that this routine
   creates can be passed through the GPFS I/O stack just like the ones
   describing buffers in the GPFS page pool. */
int cxiPinMM(struct page *pageP, cxiXmem_t* xmemDescP)
{
  struct cxiPageList_t* plP;
  int rc = 0;
  char* kaddrP;

  ENTER(0);

  /* Create a page list object and point it at the given page struct */
  kaddrP = kmap(pageP);
  plP = cxiAllocPageList(1, (unsigned long)kaddrP);
  if (plP == NULL)
  {
    kunmap(pageP);
    rc = ENOMEM;
    goto exit;
  }
  plP->plRootP = (void*)pageP;

  xmemDescP->pageListP = (void*)plP;

exit:
  EXIT(0);
  return rc;
}


/* Free a cross-memory descriptor that was created by cxiPinMM. */
void cxiUnpinMM(struct page *pageP, cxiXmem_t* xmemDescP)
{
  struct cxiPageList_t* plP;

  ENTER(0);
  kunmap(pageP);

  /* Clear pointer to page so cleanup will not decrement the page reference
     count, then free the page list object */
  plP = (struct cxiPageList_t*) xmemDescP->pageListP;
  plP->plRootP = NULL;
  cxiDeallocPageList(plP);
  EXIT(0);
}
#endif /* MMAP_DIO */


/* Fanout of a radix tree page */
#define RADIX_PAGE_FANOUT (PAGE_SIZE/sizeof(void*))

/* Allocate a cxiPageList_t and prepare it to record the addresses
   of up to maxPages pages.  The lowest user address in the set of
   pages to be recorded is baseAddr.  Returns NULL on any error. */
struct cxiPageList_t* cxiAllocPageList(int maxPages, UIntPtr baseAddr)
{
  struct cxiPageList_t* plP;
  int nMidPtrs;

  /* Allocate space for the page list */
  ENTER(0);
  plP = (struct cxiPageList_t*)kmalloc(sizeof(struct cxiPageList_t), GFP_KERNEL);
  if (plP == NULL)
    return NULL;

  /* Verify base address is on a page boundary */
  if ((baseAddr & (PAGE_SIZE-1)) != 0)
    goto errorExit;

  /* Validate max number of pages and compute number of levels in radix
     tree.  Allocate top level pointers if level >= 1. */
  if (maxPages <= 0)
    goto errorExit;
  else if (maxPages == 1)
  {
    plP->plNLevel = 0;
    plP->plRootP = NULL;
  }
  else if (maxPages <= RADIX_PAGE_FANOUT)
  {
    plP->plNLevel = 1;
    nMidPtrs = maxPages;
    plP->plRootP = (void*)kmalloc(nMidPtrs * sizeof(void*), GFP_KERNEL);
    if (plP->plRootP == NULL)
      goto errorExit;
    memset(plP->plRootP, 0, nMidPtrs * sizeof(void*));
  }
  else if (maxPages <= RADIX_PAGE_FANOUT*RADIX_PAGE_FANOUT)
  {
    plP->plNLevel = 2;
    nMidPtrs = (maxPages+RADIX_PAGE_FANOUT-1) / RADIX_PAGE_FANOUT;
    plP->plRootP = (void*)kmalloc(nMidPtrs * sizeof(void*), GFP_KERNEL);
    if (plP->plRootP == NULL)
      goto errorExit;
    memset(plP->plRootP, 0, nMidPtrs * sizeof(void*));
  }
  else
    goto errorExit;

  /* Initialize other fields and return */
  plP->plNPages = maxPages;
  plP->plBaseAddr = baseAddr;
  TRACE3(TRACE_SHARED, 4, TRCID_ALLOC_PAGE_LIST,
         "cxiAllocPageList: maxPages %d baseAddr 0x%lX plP 0x%lX\n",
         maxPages, baseAddr, plP);
  return plP;

errorExit:
  TRACE3(TRACE_SHARED, 1, TRCID_ALLOC_PAGE_LIST_ERR,
         "cxiAllocPageList: maxPages %d baseAddr 0x%lX plP 0x%lX\n",
         maxPages, baseAddr, plP);
#ifdef PAGE_ALIGN_DEBUG
  DBGASSERT(!(baseAddr & (PAGE_SIZE-1)));
#endif
  kfree(plP);
  EXIT(0);
  return NULL;
}


/* Map pages in the caller's address space at the given address and
   record the addresses of the struct page objects */
int cxiMapAndRecordPages(struct cxiPageList_t* plP, UIntPtr uaddr,
                         int nPages)
{
  unsigned long inUaddr;
  struct page ** pagePP;
  void ** midPP;
  int rc;
  int pageIndex;
  int nPagesLeft;
  int nPagesThisChunk;
  int midIndex;
  int leafIndex;

  /* Verify buffer address is on a page boundary */
  ENTER(0);
  if ((uaddr & (PAGE_SIZE-1)) != 0)
  {
    rc = EINVAL;
    goto exit;
  }

  /* Compute index of start page */
  pageIndex = (uaddr - plP->plBaseAddr) / PAGE_SIZE;
  if (pageIndex < 0  ||
      pageIndex+nPages > plP->plNPages)
  {
    rc = EINVAL;
    goto exit;
  }

  /* While there are more pages to record, get a pointer to where
     the next page pointer should be stored and map a range of pages
     starting at that point */
  inUaddr = uaddr;
  nPagesLeft = nPages;
  while (nPagesLeft > 0)
  {
    /* Compute where to store the pointers to struct page, and insure
       that all required pointer arrays have been allocated */
    switch (plP->plNLevel)
    {
      case 0:
        nPagesThisChunk = 1;
        pagePP = (struct page **)&plP->plRootP;
        break;

      case 1:
        DBGASSERT(plP->plRootP != NULL);
        nPagesThisChunk = nPagesLeft;
        pagePP = &(((struct page **)plP->plRootP)[pageIndex]);
        break;

      case 2:
        DBGASSERT(plP->plRootP != NULL);
        midIndex = pageIndex / RADIX_PAGE_FANOUT;
        leafIndex = pageIndex % RADIX_PAGE_FANOUT;
        midPP = &(((void**)plP->plRootP)[midIndex]);
        if (*midPP == NULL)
        {
          *midPP = (void*)kmalloc(RADIX_PAGE_FANOUT * sizeof(void*), GFP_KERNEL);
          if (*midPP == NULL)
          {
            if (nPages - nPagesLeft)
              (void)cxiReleaseAndForgetPages(plP, inUaddr, nPages - nPagesLeft);
            rc = ENOMEM;
            goto exit;
          }
          memset(*midPP, 0, RADIX_PAGE_FANOUT * sizeof(void*));
          TRACE2(TRACE_SHARED, 4, TRCID_RECORD_DETAIL_2,
                 "cxiMapAndRecordPages: allocated midPP 0x%lX for uaddr 0x%lX\n",
                 midPP, uaddr);
        }
        nPagesThisChunk = MIN(nPagesLeft, RADIX_PAGE_FANOUT-leafIndex);
        pagePP = &(((struct page **)*midPP)[leafIndex]);
        break;
    }

    /* Pin pages and record pointers to their struct page objects for use
       in submitting I/O requests */
    TRACE3(TRACE_SHARED, 4, TRCID_RECORD_DETAIL,
           "cxiMapAndRecordPages: mapping %d pages into 0x%lX uaddr 0x%lX\n",
           nPagesThisChunk, pagePP, uaddr);
    down_read(&current->mm->mmap_sem);
    rc = get_user_pages(current, current->mm, uaddr,
                        nPagesThisChunk, VM_WRITE, 0 /* force */,
                        pagePP, NULL);
    up_read(&current->mm->mmap_sem);
    if (rc != nPagesThisChunk)
    {
      if (nPages - nPagesLeft)
        (void)cxiReleaseAndForgetPages(plP, inUaddr, nPages - nPagesLeft);
      rc = ENOMEM;
      goto exit;
    }

    /* Update for next loop iteration */
    pageIndex += nPagesThisChunk;
    nPagesLeft -= nPagesThisChunk;
    uaddr += nPagesThisChunk * PAGE_SIZE;
  }
  rc = 0;

exit:
  TRACE4(TRACE_SHARED, 2, TRCID_CXIMAP_RECORD,
         "cxiMapAndRecordPages: plP 0x%lX uaddr 0x%lX nPages %d rc %d\n",
         plP, inUaddr, nPages, rc);
#ifdef PAGE_ALIGN_DEBUG
  DBGASSERT(!(inUaddr & (PAGE_SIZE-1)));
#endif
  EXIT(0);
  return rc;
}


/* Release pages mapped by cxiMapAndRecordPages.  Does not free any of
   the nodes of the radix tree.  Not an error to attempt to free pages
   that are not mapped. */
int cxiReleaseAndForgetPages(struct cxiPageList_t* plP, UIntPtr uaddr,
                             int nPages)
{
  struct page ** pagePP;
  struct page * pageP;
  void * midP;
  int rc;
  int pageIndex;
  int nPagesLeft;
  int midIndex;
  int leafIndex;

  ENTER(0);
  TRACE3(TRACE_SHARED, 2, TRCID_CXIFORGET_PAGES,
         "cxiReleaseAndForgetPages: plP 0x%lX uaddr 0x%lX nPages %d\n",
         plP, uaddr, nPages);

  /* Verify buffer address is on a page boundary */
  if ((uaddr & (PAGE_SIZE-1)) != 0)
  {
    rc = EINVAL;
    goto exit;
  }

  /* Compute index of start page */
  pageIndex = (uaddr - plP->plBaseAddr) / PAGE_SIZE;
  if (pageIndex < 0  ||
      pageIndex+nPages > plP->plNPages)
  {
    rc = EINVAL;
    goto exit;
  }

  /* For each page to unmap, decrement its reference count and clear the
     pointer to it */
  nPagesLeft = nPages;
  while (nPagesLeft > 0)
  {
    /* Walk page list according to its structure */
    switch (plP->plNLevel)
    {
      case 0:
        pagePP = (struct page **)&plP->plRootP;
        pageP = *pagePP;
        *pagePP = NULL;
        break;

      case 1:
        DBGASSERT(plP->plRootP != NULL);
        pagePP = &(((struct page **)plP->plRootP)[pageIndex]);
        pageP = *pagePP;
        *pagePP = NULL;
        break;

      case 2:
        DBGASSERT(plP->plRootP != NULL);
        midIndex = pageIndex / RADIX_PAGE_FANOUT;
        leafIndex = pageIndex % RADIX_PAGE_FANOUT;
        midP = ((void**)plP->plRootP)[midIndex];
        if (midP == NULL)
          pageP = NULL;
        else
        {
          pagePP = &(((struct page **)midP)[leafIndex]);
          pageP = *pagePP;
          *pagePP = NULL;
        }
        break;
    }

    /* Release the page */
    if (pageP != NULL)
    {
      TRACE3(TRACE_SHARED, 12, TRCID_FORGET_DETAIL,
             "cxiReleaseAndForgetPages: release page 0x%lX index %d uaddr 0x%lX\n",
             pageP, pageIndex, uaddr);
      page_cache_release(pageP);
    }

    /* Update for next loop iteration */
    pageIndex += 1;
    nPagesLeft -= 1;
    uaddr += PAGE_SIZE;
  }
  rc = 0;

exit:
#ifdef PAGE_ALIGN_DEBUG
  DBGASSERT(!(uaddr & (PAGE_SIZE-1)));
#endif
  EXIT(0);
  return rc;
}


/* Given a page list object and a user buffer address that is contained
   within the area described by the page list, copy the addresses of
   the struct page objects describing the buffer into the given array
   pagePP.  It is a fatal error if the user buffer is not completely
   contained within the area described by the page list. */
void cxiGetPagePtrs(struct cxiPageList_t* plP, UIntPtr uaddr,
                    int nPages, struct page ** pagePP)
{
  struct page ** srcPagePP;
  void * midP;
  struct page * pageP;
  int pageIndex;
  int nPagesLeft;
  int nPagesThisChunk;
  int midIndex;
  int leafIndex;
  int i;

  ENTER(0);
  TRACE4(TRACE_SHARED, 5, TRCID_GET_PAGEPTRS,
         "cxiGetPagePtrs: plP 0x%lX uaddr 0x%lX nPages %d pagePP 0x%lX\n",
         plP, uaddr, nPages, pagePP);

  /* Validate parameters */
  LOGASSERT(uaddr >= plP->plBaseAddr);
  pageIndex = (uaddr - plP->plBaseAddr) / PAGE_SIZE;
  LOGASSERT(pageIndex+nPages <= plP->plNPages);
  nPagesLeft = nPages;
  while (nPagesLeft > 0)
  {
    /* Compute address of next saved struct page pointer */
    switch (plP->plNLevel)
    {
      case 0:
        nPagesThisChunk = 1;
        srcPagePP = (struct page **)&plP->plRootP;
        break;

      case 1:
        LOGASSERT(plP->plRootP != NULL);
        nPagesThisChunk = nPagesLeft;
        srcPagePP = &(((struct page **)plP->plRootP)[pageIndex]);
        break;

      case 2:
        LOGASSERT(plP->plRootP != NULL);
        midIndex = pageIndex / RADIX_PAGE_FANOUT;
        leafIndex = pageIndex % RADIX_PAGE_FANOUT;
        midP = ((void**)plP->plRootP)[midIndex];
        LOGASSERT(midP != NULL);
        nPagesThisChunk = MIN(nPagesLeft, RADIX_PAGE_FANOUT-leafIndex);
        srcPagePP = &(((struct page **)midP)[leafIndex]);
        break;
    }

    /* Copy page pointers to target array */
    for (i=0; i<nPagesThisChunk; i++)
    {
      pageP = srcPagePP[i];
      TRACE4(TRACE_SHARED, 12, TRCID_GET_PAGEPTRS_DETAILS,
             "cxiGetPagePtrs: plP 0x%lX i %d/%d pageP 0x%lX\n",
             plP, i, nPagesThisChunk, pageP);
      pagePP[i] = pageP;
    }

    /* Update for next loop iteration */
    pagePP += nPagesThisChunk;
    pageIndex += nPagesThisChunk;
    nPagesLeft -= nPagesThisChunk;
  }
  EXIT(0);
}


/* Release all pages covered by the given page list, then free all kernel
   storage associated with managing the list */
void cxiDeallocPageList(struct cxiPageList_t* plP)
{
  int nMidPtrs;
  int i;
  void * midP;

  TRACE1(TRACE_SHARED, 2, TRCID_DEALLOC_PAGE_LIST,
         "cxiDeallocPageList: plP 0x%lX\n", plP);

  /* Release all pages on this page list */
  ENTER(0);
  (void)cxiReleaseAndForgetPages(plP, plP->plBaseAddr, plP->plNPages);

  /* Free all levels of the radix tree */
  switch (plP->plNLevel)
  {
    case 0:
      break;

    case 1:
      DBGASSERT(plP->plRootP != NULL);
      kfree(plP->plRootP);
      plP->plRootP = NULL;
      break;

    case 2:
      nMidPtrs = (plP->plNPages+RADIX_PAGE_FANOUT-1) / RADIX_PAGE_FANOUT;
      for (i=0; i<nMidPtrs; i++)
      {
        midP = ((void**)plP->plRootP)[i];
        if (midP != NULL)
          kfree(midP);
      }
      kfree(plP->plRootP);
      plP->plRootP = NULL;
      break;
  }

  /* Free the cxiPageList_t itself */
  plP->plNPages = -1;
  kfree(plP);
  EXIT(0);
}


/* Attach an I/O buffer to the kernel's virtual address space.  The
   cxiIOBufferAttachment_t returned in *attachP must be used as a parameter of
   most of the other operations on cxiIOBuffer_t's. */
void 
cxiAttachIOBuffer(struct cxiIOBuffer_t* iobP,
                  cxiXmem_t* xmemDescP,
                  struct cxiIOBufferAttachment_t* attachP)
{
  int rc;

  /* Use the cxiPageList_t as the attachment data. */
  attachP->pageListP = (struct cxiPageList_t*)xmemDescP->pageListP;
  DBGASSERT(attachP->pageListP != NULL);
  TRACE1(TRACE_KSVFS, 11, TRCID_ATTACH_KIBD,
         "cxiAttachIOBuffer: plP 0x%lX\n", attachP->pageListP);
  EXIT(0);
}


/* Detach a buffer from the kernel's virtual address space. */
void 
cxiDetachIOBuffer(struct cxiIOBuffer_t* iobP,
                  struct cxiIOBufferAttachment_t* attachP)
{
  /* Validate attachment data */
  ENTER(0);
  TRACE2(TRACE_KSVFS, 5, TRCID_DETACH_KIBD,
         "cxiDetachIOBuffer: dataPtr 0x%lX plP 0x%lX\n",
         iobP->daemonBufP, attachP->pageListP);
  if (attachP->pageListP == NULL)
  {
    EXIT(0);
    return;
  }

  /* Invalidate attachment data */
  attachP->pageListP = NULL;
  EXIT(0);
}


/* Transfer len bytes beginning at offset bufOffset within I/O buffer *iobP
   to or from a user buffer.  The direction of the transfer is given with
   respect to the I/O buffer.  Returns EOK if successful, other error
   codes if unsuccessful. */
IntRC 
cxiUXfer(struct cxiIOBuffer_t* iobP, Boolean toIOBuffer,
         const struct cxiIOBufferAttachment_t* attachP,
         void* vkopP, int bufOffset, int len, struct cxiUio_t* uioP)
{
  char* dataAddrP;
  struct cxiPageList_t* plP = attachP->pageListP;
  unsigned long pageOffset;
  struct page * pageP;
  int pageLen;
  unsigned long kaddr;
  int rc = 0;

  ENTER(0);
  /* Validate parameters */
  TRACE5(TRACE_KSVFS, 5, TRCID_UXFER_LINUX,
         "cxiUXfer: dataPtr 0x%lX plP 0x%lX toIOBuf %d offset %d len %d\n",
         iobP->daemonBufP, plP, toIOBuffer, bufOffset, len);

  DBGASSERT(bufOffset >= 0);
  DBGASSERT(bufOffset+len <= iobP->ioBufLen);
  DBGASSERT(iobP->ioBufLen/PAGE_SIZE <= plP->plNPages);

  /* Transfer data in or out of as many pages as necessary to satisfy
     the data move request */
  dataAddrP = iobP->daemonBufP + bufOffset;
  pageOffset = (unsigned long)dataAddrP % PAGE_SIZE;
  pageLen = PAGE_SIZE - pageOffset;
  for (;;)
  {
    /* Calculate how many bytes to move in or out of the current page of the
       I/O buffer */
    if (len < pageLen)
      pageLen = len;

    /* Map current I/O buffer page into the kernel's address space
       temporarily, then copy data in or out of the page */
    cxiGetPagePtrs(plP, (unsigned long)dataAddrP, 1, &pageP);
    kaddr = (unsigned long)kmap(pageP);
    TRACE4(TRACE_KSVFS, 12, TRCID_UXFER_UIOMOVE,
           "cxiUXfer: uiomove kaddr 0x%lX pageOffset %d pageLen %d pageP 0x%lX\n",
           kaddr, pageOffset, pageLen, pageP);

    rc = cxiUiomove((char *)(kaddr + pageOffset), pageLen, toIOBuffer, uioP);
    kunmap(pageP);

    /* Leave loop if an error occurred on the move */
    if (rc != 0)
      break;

    /* Update length left to copy and test for loop termination */
    len -= pageLen;
    if (len <= 0)
      break;

    /* Set up for next iteration */
    pageOffset = 0;
    dataAddrP += pageLen;
    pageLen = PAGE_SIZE;
  }  /* end of do forever */

  EXIT(0);
  return rc;
}


/* Perform cross-memory transfer of len bytes from user memory in current
   task to memory in specified address space.  If toXmem is true then
   copy is from userAddrP to udataP/xmemP, otherwise the opposite. */
IntRC 
cxiXmemXfer(char *userAddrP, int len, char *xdataP, cxiXmem_t* xmemP,
            Boolean toXmem)
{
  int rc = 0;
  int pageOffset, pageLen;
  void *kaddrP;
  struct page *pageP;
  struct cxiPageList_t* plP = (struct cxiPageList_t*)xmemP->pageListP;

  ENTER(0);
  TRACE5(TRACE_KSVFS, 5, TRCID_XMEMXFER_LINUX,
         "cxiXmemXfer: userAddrP 0x%lX len %d xdataP 0x%lX "
         "plP 0x%lX toXmem %d\n",
         userAddrP, len, xdataP, plP, toXmem);

  /* Transfer data in or out of as many pages as necessary to satisfy
     the data move request */
  pageOffset = ((unsigned long)xdataP - plP->plBaseAddr) % PAGE_SIZE;
  pageLen = PAGE_SIZE - pageOffset;
  for (;;)
  {
    /* Calculate how many bytes to move in or out of the current page of the
       I/O buffer */
    if (len < pageLen)
      pageLen = len;

    /* Map current I/O buffer page into the kernel's address space
       temporarily, then copy data in or out of the page */
    cxiGetPagePtrs(plP, (unsigned long)xdataP, 1, &pageP);
    kaddrP = kmap(pageP);
    TRACE4(TRACE_KSVFS, 12, TRCID_XMEMFER_COPY,
           "cxiXmemXfer: copy kaddrP 0x%lX pageOffset %d pageLen %d pageP 0x%lX\n",
           kaddrP, pageOffset, pageLen, pageP);

    if (toXmem)
      rc = cxiCopyIn(userAddrP, (char *)kaddrP + pageOffset, pageLen);
    else
      rc = cxiCopyOut((char *)kaddrP + pageOffset, userAddrP, pageLen);

    kunmap(pageP);

    /* Leave loop if an error occurred on the move */
    if (rc != 0)
      break;

    /* Update length left to copy and test for loop termination */
    len -= pageLen;
    if (len <= 0)
      break;

    /* Set up for next iteration */
    userAddrP += pageLen;
    xdataP += pageLen;
    pageOffset = 0;
    pageLen = PAGE_SIZE;
  }  /* end of do forever */

  EXIT(0);
  return rc;
}


/* Transfer len bytes beginning at offset bufOffset within I/O buffer *iobP
   to or from a contiguous kernel buffer.  The direction of the transfer
   is given with respect to the I/O buffer.  Returns EOK if successful,
   other error codes if unsuccessful. */
IntRC 
cxiKXfer(struct cxiIOBuffer_t* iobP, Boolean toIOBuffer,
         const struct cxiIOBufferAttachment_t* attachP,
         int bufOffset, int len, char* kBufP)
{
  struct cxiPageList_t* plP = attachP->pageListP;
  char* dataAddrP;
  unsigned long pageOffset;
  struct page * pageP;
  int pageLen;
  unsigned long kaddr;

  /* Validate parameters */
  ENTER(0);
  TRACE6(TRACE_KSVFS, 5, TRCID_KXFER_LINUX,
         "cxiKXfer: dataPtr 0x%lX plP 0x%lX toIOBuf %d offset %d len %d kBufP 0x%lX\n",
         iobP->daemonBufP, plP, toIOBuffer, bufOffset, len, kBufP);

  DBGASSERT(bufOffset >= 0);
  DBGASSERT(bufOffset+len <= iobP->ioBufLen);
  DBGASSERT(iobP->ioBufLen/PAGE_SIZE <= plP->plNPages);

  /* Transfer data in or out of as many pages as necessary to satisfy
     the data move request */
  dataAddrP = iobP->daemonBufP + bufOffset;
  pageOffset = (unsigned long)dataAddrP % PAGE_SIZE;
  pageLen = PAGE_SIZE - pageOffset;
  for (;;)
  {
    /* Calculate how many bytes to move in or out of the current page of the
       I/O buffer */
    if (len < pageLen)
      pageLen = len;

    /* Map current I/O buffer page into the kernel's address space
       temporarily, then copy data in or out of the page */
    cxiGetPagePtrs(plP, (unsigned long)dataAddrP, 1, &pageP);
    kaddr = (unsigned long)kmap(pageP);
    TRACE4(TRACE_KSVFS, 12, TRCID_KXFER_MEMCPY,
           "cxiKXfer: move kaddr 0x%lX pageOffset %d pageLen %d pageP 0x%lX\n",
           kaddr, pageOffset, pageLen, pageP);

    if (toIOBuffer)
      memcpy((void *)(kaddr + pageOffset), kBufP, pageLen);
    else
      memcpy(kBufP, (void *)(kaddr + pageOffset), pageLen);
    kunmap(pageP);

    /* Update length left to copy and test for loop termination */
    len -= pageLen;
    if (len <= 0)
      break;

    /* Set up for next iteration */
    dataAddrP += pageLen;
    kBufP += pageLen;
    pageOffset = 0;
    pageLen = PAGE_SIZE;
  }  /* end of do forever */

  EXIT(0);
  return 0;
}


/* Set len bytes beginning at offset bufOffset within I/O buffer *iobP
   to zero.  Returns EOK if successful, other error codes if unsuccessful. */
IntRC 
cxiKZero(struct cxiIOBuffer_t* iobP,
         const struct cxiIOBufferAttachment_t* attachP,
         int bufOffset, int len)
{
  struct cxiPageList_t* plP = attachP->pageListP;
  char* dataAddrP;
  unsigned long pageOffset;
  struct page * pageP;
  int pageLen;
  unsigned long kaddr;

  /* Validate parameters */
  ENTER(0);
  TRACE4(TRACE_KSVFS, 5, TRCID_KZERO_LINUX,
         "cxiKZero: dataPtr 0x%lX plP 0x%lX offset %d len %d\n",
         iobP->daemonBufP, plP, bufOffset, len);

  DBGASSERT(bufOffset >= 0);
  DBGASSERT(bufOffset+len <= iobP->ioBufLen);
  DBGASSERT(iobP->ioBufLen/PAGE_SIZE <= plP->plNPages);

  /* Zero data in as many pages as necessary to complete the request */
  dataAddrP = iobP->daemonBufP + bufOffset;
  pageOffset = (unsigned long)dataAddrP % PAGE_SIZE;
  pageLen = PAGE_SIZE - pageOffset;
  for (;;)
  {
    /* Calculate how many bytes to zero in the current page of the I/O
       buffer */
    if (len < pageLen)
      pageLen = len;

    /* Map current I/O buffer page into the kernel's address space
       temporarily, then zero data in the page */
    cxiGetPagePtrs(plP, (unsigned long)dataAddrP, 1, &pageP);
    kaddr = (unsigned long)kmap(pageP);
    TRACE4(TRACE_KSVFS, 12, TRCID_KZERO_MEMSET,
           "cxiKZero: zero kaddr 0x%lX pageOffset %d pageLen %d pageP 0x%lX\n",
           kaddr, pageOffset, pageLen, pageP);
    memset((void *)(kaddr + pageOffset), 0, pageLen);
    kunmap(pageP);

    /* Update length left to zero and test for loop termination */
    len -= pageLen;
    if (len <= 0)
      break;

    /* Set up for next iteration */
    dataAddrP += pageLen;
    pageOffset = 0;
    pageLen = PAGE_SIZE;
  }  /* end of do forever */

  EXIT(0);
  return 0;
}


/* Map an I/O buffer so it can be read and written from kernel code
   running in the context of a user thread.  Depending on the platform, the
   addresses at which the I/O buffer gets mapped may not be contiguous.  The
   details of how the buffer got mapped are handled by the
   cxiDiscontiguousDirectoryBuffer_t object that is filled in by this call.
   On some platforms, mapping buffers using this call consumes scarce
   resources, so all cxiMapDiscontiguousRW calls should be promptly matched by
   cxiUnmapDiscontiguousRW calls as soon as the operation that required access
   to the I/O buffer completes.  Returns 0 if successful, other error codes
   if unsuccessful. */
IntRC 
cxiMapDiscontiguousRW(struct cxiIOBuffer_t* iobP,
                      const struct cxiIOBufferAttachment_t* attachP,
                      struct cxiDiscontiguousDirectoryBuffer_t* discontigP)
{
  /* ?? WARNING: Since this must kmap multiple pages, there is the
     possibility of deadlock if multiple threads are part of the way through
     executing this code, and LAST_PKMAP pages (512 or 1024) have already
     been kmapped.  There needs to be flow control whereby threads reserve
     enough pages to complete all of their kmaps before they begin acquiring
     pages. */
  struct cxiPageList_t* plP = attachP->pageListP;
  char* dataAddrP;
  int dirIndex;
  int mapPages;
  struct page * pageP;
  unsigned long kaddr;
  unsigned long pageoffSet;

  /* __CXI_BUFFERS_ARE_CONTIGUOUS is not #defined */

  /* Validate parameters */
  ENTER(0);
  TRACE3(TRACE_KSVFS, 4, TRCID_MAP_DISCONTIG_ENTER,
         "cxiMapDiscontiguousRW: dataPtr 0x%lX plP 0x%lX ioBufLen 0x%X\n",
         iobP->daemonBufP, plP, iobP->ioBufLen);

  /* The mappable buffer memory may be longer than a directory block */
  mapPages = (iobP->ioBufLen + DISCONTIG_PAGE_SIZE - 1) / DISCONTIG_PAGE_SIZE;
  mapPages = MIN(mapPages, MAX_PAGES_PER_DIRBLOCK);

  pageoffSet = (unsigned long)iobP->daemonBufP % DISCONTIG_PAGE_SIZE;
  dataAddrP = iobP->daemonBufP - pageoffSet;
  /* Must be page aligned if more than 1 page */
  DBGASSERT(mapPages == 1 || pageoffSet == 0);
  for (dirIndex=0 ; dirIndex<mapPages ; dirIndex++)
  {
    cxiGetPagePtrs(plP, (unsigned long)dataAddrP, 1, &pageP);
    if (pageP == NULL)
      break;

    kaddr = (unsigned long)kmap(pageP);
    TRACE4(TRACE_KSVFS, 12, TRCID_MAP_DISCONTIG_KMAP,
           "cxiMapDiscontiguousRW: dirIndex %d pageP 0x%lX kaddr 0x%lX pageoffSet 0x%lX\n",
           dirIndex, pageP, kaddr, pageoffSet);

    DBGASSERT(dirIndex < MAX_PAGES_PER_DIRBLOCK);
    discontigP->userPagePointerArray[dirIndex] = (char*)kaddr + pageoffSet;
    discontigP->osPagePointerArray[dirIndex] = (void*)pageP;

    dataAddrP += PAGE_SIZE;
  }

  discontigP->mappedLen = dirIndex * DISCONTIG_PAGE_SIZE;
  EXIT(0);
  return 0;
}


/* Unmap an I/O buffer previously mapped */
void 
cxiUnmapDiscontiguousRW(struct cxiIOBuffer_t* iobP,
                        struct cxiDiscontiguousDirectoryBuffer_t* discontigP)
{
  int checkPageIndex;
  struct page * pageP;
  int mappedPages;

  ENTER(0);
  TRACE3(TRACE_KSVFS, 4, TRCID_UNMAP_DISCONTIG_ENTER,
         "cxiUnmapDiscontiguousRW: dataPtr 0x%lX ioBufLen 0x%X mappedLen %d\n",
         iobP->daemonBufP, iobP->ioBufLen, discontigP->mappedLen);

  /* Unmap all pages in discontiguous map.  If the osPagePointerArray entry
   * is NULL, it means that the last mapping was made via MapContiguousBuffer,
   * which did not do any kmaps that need to be kunmap'ped. 
   */
  mappedPages = (discontigP->mappedLen + DISCONTIG_PAGE_SIZE - 1) /
                DISCONTIG_PAGE_SIZE;

  for (checkPageIndex = 0; checkPageIndex < mappedPages; checkPageIndex++)
  {
    pageP = (struct page *)discontigP->osPagePointerArray[checkPageIndex];
    TRACE3(TRACE_KSVFS, 12, TRCID_UNMAP_DISCONTIG_KUNMAP,
           "cxiUnmapDiscontiguousRW: unmap checkPageIndex %d pageP 0x%lX "
           "kaddr 0x%lX\n", checkPageIndex, pageP,
           discontigP->userPagePointerArray[checkPageIndex]);

    if (pageP != NULL)
    {
      kunmap(pageP);
      discontigP->osPagePointerArray[checkPageIndex] = NULL;
    }
    discontigP->userPagePointerArray[checkPageIndex] = NULL;
  }
  discontigP->mappedLen = 0;
  EXIT(0);
}

/* Return an address in kernel memory that holds a contigous read-only
   copy of a portion of an I/O buffer.  If possible, this will be a
   mapping of the I/O buffer.  If necessary, this routine will allocate a
   new block of kernel memory and copy the requested data to it.  The
   returned cxiContiguousBuffer_t encapsulates what method was used, so
   that cxiUnmapContiguousRO can release whatever resources were obtained by
   this call.  Returns 0 if successful, other error codes if
   unsuccessful. */
IntRC 
cxiMapContiguousRO(struct cxiIOBuffer_t* iobP,
                   const struct cxiIOBufferAttachment_t* attachP,
                   int bufOffset, int len, const char** contigBasePP,
                   struct cxiContiguousBuffer_t* contigP)
{
  int startPageIndex;
  unsigned long pageOffset;
  char* dataAddrP;
  int endPageIndex;
  struct cxiPageList_t* plP = attachP->pageListP;
  struct page * pageP;
  unsigned long kaddr;
  char* tempBufP;
  Boolean usedKmalloc;
  int rc;

  /* Validate parameters */
  ENTER(0);
  TRACE5(TRACE_KSVFS, 4, TRCID_MAP_CONTIG_ENTER,
         "cxiMapContiguousRO: dataPtr 0x%lX ioBufLen %d plP 0x%lX bufOffset %d len %d\n",
         iobP->daemonBufP, iobP->ioBufLen, plP, bufOffset, len);

  DBGASSERT(bufOffset >= 0);
  DBGASSERT(bufOffset+len <= iobP->ioBufLen);
  DBGASSERT(iobP->ioBufLen/PAGE_SIZE <= plP->plNPages);

  /* If the requested piece of the I/O buffer does not cross a page boundary,
     then map the page and return the mapped address within the page */
  startPageIndex = bufOffset / PAGE_SIZE;
  dataAddrP = iobP->daemonBufP + bufOffset;
  pageOffset = (unsigned long)dataAddrP % PAGE_SIZE;
  endPageIndex = (bufOffset+len-1) / PAGE_SIZE;

  TRACE4(TRACE_KSVFS, 5, TRCID_MAP_CONTIG_ENTER_2,
         "cxiMapContiguousRO: startPageIndex %d endPageIndex %d pageOffset 0x%lX dataAddrP 0x%lX\n",
         startPageIndex, endPageIndex, pageOffset, dataAddrP);

  if (startPageIndex == endPageIndex)
  {
    /* Map I/O buffer page into the kernel's address space */
    cxiGetPagePtrs(plP, (unsigned long)dataAddrP, 1, &pageP);
    kaddr = (unsigned long)kmap(pageP);

    /* Return address within the mapped page, and set map state so
       cxiUnmapContiguousRO knows to do kunmap */
    *contigBasePP = (char*) (kaddr+pageOffset);
    contigP->mallocedBaseP = NULL;
    contigP->usedKmalloc = false;
    contigP->pageP = pageP;
    TRACE2(TRACE_KSVFS, 5, TRCID_MAP_CONTIG_KMAP,
           "cxiMapContiguousRO: mapped pageP 0x%lX at 0x%lX\n",
           pageP, *contigBasePP);
    EXIT(0);
    return 0;
  }

  /* Otherwise, the requested part of the I/O buffer spans page boundaries.
     Allocate a contiguous buffer, and copy data from the I/O buffer to the
     temporary buffer. */
  else
  {
    if (len <= PAGE_SIZE)
    {
      tempBufP = (char *)kmalloc(len, GFP_KERNEL);
      usedKmalloc = true;
    }
    else
    {
      tempBufP = (char*)vmalloc(len);
      usedKmalloc = false;
    }
    if (tempBufP == NULL)
    {
      EXIT(0);
      return -ENOMEM;
    }
    rc = cxiKXfer(iobP, CXI_XFER_FROM_IOBUFFER, attachP, bufOffset, len,
                  tempBufP);
    if (rc != 0)
    {
      if (usedKmalloc)
        kfree((void*)tempBufP);
      else
        vfree((void*)tempBufP);
      EXIT(0);
      return rc;
    }
#ifdef MALLOC_DEBUG
    MallocDebugNew(tempBufP, len, 4);
#endif

    /* Return address within the contiguous temporary buffer, and set map
       state so cxiUnmapContiguousRO knows to do vfree */
    *contigBasePP = tempBufP;
    contigP->mallocedBaseP = tempBufP;
    contigP->usedKmalloc = usedKmalloc;
    contigP->pageP = NULL;
    TRACE1(TRACE_KSVFS, 5, TRCID_MAP_CONTIG_VMALLOC,
           "cxiMapContiguousRO: copied to 0x%lX\n", tempBufP);
    EXIT(0);
    return 0;
  }
}


/* Release a mapping or copy obtained with cxiMapContiguousRO */
void 
cxiUnmapContiguousRO(struct cxiIOBuffer_t* iobP,
                     struct cxiContiguousBuffer_t* contigP)
{
  ENTER(0);
  if (contigP->mallocedBaseP != NULL)
  {
    TRACE2(TRACE_KSVFS, 4, TRCID_UNMAP_CONTIG_VFREE,
           "cxiUnmapContiguousRO: dataPtr 0x%lX kfree/vfree 0x%lX\n",
           iobP->daemonBufP, contigP->mallocedBaseP);
    DBGASSERT(contigP->pageP == NULL);

    if (contigP->usedKmalloc)
      kfree((void*)contigP->mallocedBaseP);
    else
      vfree((void*)contigP->mallocedBaseP);

#ifdef MALLOC_DEBUG
    MallocDebugDelete(contigP->mallocedBaseP);
#endif
    contigP->mallocedBaseP = NULL;
  }
  else
  {
    TRACE2(TRACE_KSVFS, 4, TRCID_UNMAP_CONTIG_KUNMAP,
           "cxiUnmapContiguousRO: dataPtr 0x%lX kunmap 0x%lX\n",
           iobP->daemonBufP, contigP->pageP);
    DBGASSERT(contigP->pageP != NULL);
    kunmap((struct page *)contigP->pageP);
    contigP->pageP = NULL;
  }
  EXIT(0);
}


/* Routine to call when the last I/O in a chunk list has finished.
   Reports completion and error status to higher-level layer and wakes
   up any waiting thread.  It is possible for this routine to be called
   twice: if the last I/O completion interrupt occurs just after the
   extra increment of nIODoneLeft is undone near the end of cxiStartIO,
   then both the iodone handler and cxiStartIO will observe nIODoneLeft
   as zero and will call this routine.  Decide whether this routine has
   already been called by examining bhcDoneLink to see if the chunk list
   has already been enqueued on a completion queue. */
static void chunkListDone(struct cxiBufHeadChunk_t* bhcHeadP)
{
  struct cxiIODoneQueue_t*  queueP;
  struct cxiBufHeadChunk_t*  bhcP;
  unsigned long flags;

  /* Exit without doing anything if this routine has already been called */
  queueP = bhcHeadP->bhcCompletionQueueP;
  SPIN_LOCK_IRQ(&queueP->ioqWaitQ.lock, flags);
  if (bhcHeadP->bhcDoneLink.next != &bhcHeadP->bhcDoneLink)
  {
    SPIN_UNLOCK_IRQ(&queueP->ioqWaitQ.lock, flags);
    return;
  }

  /* Callback to inform higher level that I/O is complete.  Must be done
     before a call to cxiWaitIO finds that it does not need to block. */
  bhcHeadP->bhcIntCallbackRtn(bhcHeadP->bhcCallbackDataP);

  /* Add this chunk to the list of completed I/Os */
  list_add_tail(&bhcHeadP->bhcDoneLink, &queueP->ioqListHead);
  atomic_inc(&queueP->ioqNQueued);
  SPIN_UNLOCK_IRQ(&queueP->ioqWaitQ.lock, flags);

  /* Wake up waiter in cxiWaitIO if there is one and this completion
     satisfies the wait */
  if (atomic_read(&queueP->ioqNQueued) >= atomic_read(&queueP->ioqNDesired))
    wake_up(&queueP->ioqWaitQ);
}


#if LINUX_KERNEL_VERSION < 2050000
/* iodone routine for GPFS buffer_heads (Linux 2.4) */
static void 
bufferHeadDone(struct buffer_head* bhP, int uptodate)
{
  struct cxiBufHeadChunk_t* bhcHeadP;

  /* Update buffer state */
  mark_buffer_uptodate(bhP, uptodate);
  unlock_buffer(bhP);

  /* Decrement number of outstanding buffer_heads */
  bhcHeadP = (struct cxiBufHeadChunk_t*)bhP->b_private;
  atomic_dec(&bhcHeadP->nIODoneLeft);

  /* If this is the last bio in the chunk list, enqueue the chunk list
     on the completion queue */
  if (atomic_read(&bhcHeadP->nIODoneLeft) == 0)
    chunkListDone(bhcHeadP);
}


/* Start a read or write of the given sectors from dev.  Data should be
   placed into the I/O buffer beginning at address userAddrP.  Returns 0
   on success, errno values on error. */
/* Linux 2.4 version */
int
cxiStartIO(int ioVecSize,
           cxiIOVecEntry* iovP,
           Boolean isWrite, cxiDev_t dev,
           UInt64 startSector, int sectorSize,
           void* qP,
           cxiIODoneIntCallback_t intCallbackRtn,
           void* callbackParmP)
{
  int nTotalPages;
  struct cxiBufHeadChunk_t* bhcP;
  struct cxiBufHeadChunk_t* bhcHeadP;
  struct cxiBufHeadChunk_t* bhcTailP;
  int nBHsAllocated;
  int pageOffset;
  int sectorsThisBH;
  unsigned long bufOffset;
  unsigned long bufEndOffset;
  struct buffer_head* bhP;
  struct page* pageP;
  struct cxiBufHeadChunk_t* p;
  struct cxiIODoneQueue_t* queueP = (struct cxiIODoneQueue_t*)qP;
  kdev_t kdev = cxiDevToKernelDev(dev); /* Convert to kernel version of dev_t */

  cxiXmem_t* xmemDescP = iovP[0].xmemDescP;
  char* userAddrP = iovP[0].baseEffAddr;
  int nSectors = iovP[0].nSectors;

  ENTER(0);
  TRACE7(TRACE_IO, 5, TRCID_KDOIO_LINUX,
         "cxiStartIO: xmemP 0x%lX uaddr 0x%lX isWrite %d dev 0x%X sector %llu nSectors %d queueP 0x%lX\n",
         xmemDescP, userAddrP, isWrite, dev, startSector, nSectors,
         queueP);

  /* Validate parameters */
  LOGASSERT(ioVecSize == 1);
  DBGASSERT(nSectors > 0);

  /* Compute the total number of pages spanned by the portion of the
     buffer that will participate in the I/O.  This equals the number
     of buffer_heads that will be used. */
  bufOffset = userAddrP -
              (char*)(((struct cxiPageList_t*)(xmemDescP->pageListP))->plBaseAddr);
  bufEndOffset = bufOffset + nSectors*512 - 1;
  nTotalPages = (bufEndOffset/PAGE_SIZE) - (bufOffset/PAGE_SIZE) + 1;
  pageOffset = bufOffset % PAGE_SIZE;
  DBGASSERT(pageOffset%512 == 0);

  /* Allocate the first buffer_head chunk needed for this I/O */
  bhcP = (struct cxiBufHeadChunk_t*) kmem_cache_alloc(BhcCacheP, SLAB_KERNEL);
  bhcHeadP = bhcP;
  bhcTailP = bhcP;
  if (bhcP == NULL)
    goto enomem;
  INIT_LIST_HEAD(&bhcP->bhcDoneLink);
  bhcP->bhcCompletionQueueP = queueP;
  bhcP->bhcIntCallbackRtn = intCallbackRtn;
  bhcP->bhcCallbackDataP = callbackParmP;
  bhcP->bhcNextP = bhcP; /* circular link to itself */
  bhcP->nSlotsUsed = 0;

  /* Initialize number of I/Os remaining to 1 to prevent bufferHeadDone from
     deciding that the I/O is completely done before all of its chunks have
     been submitted. */
  atomic_set(&bhcP->nIODoneLeft, 1);

  /* Allocate the remainder of the list of buffer_head chunks needed for
     this I/O */
  nBHsAllocated = BUFFER_HEADS_PER_CHUNK;
  while (nBHsAllocated < nTotalPages)
  {
    bhcP = (struct cxiBufHeadChunk_t*) kmem_cache_alloc(BhcCacheP, SLAB_KERNEL);
    if (bhcP == NULL)
      goto enomem;
    bhcP->nSlotsUsed = 0;

    bhcP->bhcNextP = bhcHeadP;
    bhcTailP->bhcNextP = bhcP;
    bhcTailP = bhcP;
    nBHsAllocated += BUFFER_HEADS_PER_CHUNK;
  }

  /* Build and submit a buffer_head for each page of the current I/O */
  sectorsThisBH = MIN((PAGE_SIZE-pageOffset) / 512, nSectors);
  bhcP = bhcHeadP;
  while (nSectors > 0)
  {
    /* Get a buffer_head for the the next page */
    if (bhcP->nSlotsUsed == BUFFER_HEADS_PER_CHUNK)
    {
      bhcP = bhcP->bhcNextP;
      DBGASSERT(bhcP->nSlotsUsed == 0);
    }
    bhP = &bhcP->bh[bhcP->nSlotsUsed];
    bhcP->nSlotsUsed += 1;

    /* Initialize the new buffer_head */
    memset(bhP, 0, sizeof(*bhP));

    cxiGetPagePtrs((struct cxiPageList_t*)(xmemDescP->pageListP),
                   (unsigned long)userAddrP, 1, &pageP);
    DBGASSERT(pageP != NULL);
    userAddrP += (sectorsThisBH * 512);

    /* Build and submit the buffer_head for the current page */
    bhP->b_size = sectorsThisBH * 512;
    bhP->b_page = pageP;
#ifndef __64BIT__
    if (PageHighMem(pageP))
      bhP->b_data = (char *)(0 + pageOffset);
    else
#endif
      bhP->b_data = page_address(pageP) + pageOffset;

    bhP->b_this_page = bhP;
    bhP->b_end_io = bufferHeadDone;
    bhP->b_private = (void*)bhcHeadP;
    bhP->b_blocknr = startSector;
    init_waitqueue_head(&bhP->b_wait);
    bhP->b_dev = kdev;
    bhP->b_rdev = kdev;
    bhP->b_rsector = startSector;
    bhP->b_list = BUF_CLEAN;
    bhP->b_state = (1 << BH_Mapped) | (1 << BH_Lock) | (1 << BH_Req) |
                   (1 << BH_Uptodate);
    TRACE6(TRACE_IO, 6, TRCID_KDOIO_LINUX_BH,
           "cxiStartIO: bhcP 0x%lX bhP 0x%lX sector %llu sectorsThisBH %d state 0x%X pageP 0x%lX\n",
           bhcP, bhP, startSector, sectorsThisBH, bhP->b_state, pageP);

    atomic_inc(&bhcHeadP->nIODoneLeft);
    generic_make_request(isWrite, bhP);

    if (isWrite)
      MOD_PGPGOUT(sectorsThisBH);
    else
      MOD_PGPGIN(sectorsThisBH);

    /* Advance to next page */
    startSector += sectorsThisBH;
    nSectors -= sectorsThisBH;
    sectorsThisBH = MIN(nSectors, PAGE_SIZE/512);
    pageOffset = 0;
  }

  /* Unplug the disk to be sure I/Os actually get started */
  run_task_queue(&tq_disk);

  /* Remove the extra count put in bhcHeadP->nIODoneLeft at the
     beginning of this routine.  If the count is now zero, perform chunk
     done handling */
  atomic_dec(&bhcHeadP->nIODoneLeft);
  if (atomic_read(&bhcHeadP->nIODoneLeft) == 0)
    chunkListDone(bhcHeadP);

  EXIT(0);
  return 0;

enomem:

  /* Free buffer_head chunks allocated so far and return failure */
  if (bhcHeadP != NULL)
  {
    bhcP = bhcHeadP;
    do
    {
      p = bhcP;
      bhcP = bhcP->bhcNextP;
      kmem_cache_free(BhcCacheP, (void*)p);
    }
    while (p != bhcTailP);
  }
  EXIT(0);
  return ENOMEM;
}


/* Routine to set up the disk block size and get disk parameters */
int 
GetDiskInfoX(cxiDev_t devId, struct cxiDiskInfo_t* diskInfoP)
{
  kdev_t kdev;
  unsigned int n1KBlocks;

  /* Convert to kernel version of dev_t */
  ENTER(0);
  kdev = cxiDevToKernelDev(devId);

  /* Get hardware sector size.  If unknown, assume 512. */
#if LINUX_KERNEL_VERSION >= 2040312
  diskInfoP->sectorSize = get_hardsect_size(kdev);
#else
  diskInfoP->sectorSize = get_hardblocksize(kdev);
#endif
  if (diskInfoP->sectorSize == 0)
    diskInfoP->sectorSize = 512;

  /* Set blocksize of this device to hardware sector size */
  set_blocksize(kdev, diskInfoP->sectorSize);

  /* If defined, return number of sectors on device */
  n1KBlocks = 0;
  if (blk_size[MAJOR(kdev)])
    n1KBlocks = blk_size[MAJOR(kdev)][MINOR(kdev)];
  diskInfoP->totalSectors = (Int64)n1KBlocks * 1024 / diskInfoP->sectorSize;
  TRACE3(TRACE_IO, 2, TRCID_DISKINFO,
         "GetDiskInfo: devId %08llX sector size %d totalSectors %lld\n",
         devId, diskInfoP->sectorSize, diskInfoP->totalSectors);
#if 0
  printk("VMALLOC_START=0x%lX VMALLOC_END=0x%lX\n",
         VMALLOC_START, VMALLOC_END);
#endif

  EXIT(0);
  return 0;
}

#else /* >= 2050000 */

/* iodone routine for struct bio (Linux 2.6) */
static int 
bioDone(struct bio* bioP, unsigned int done, int err)
{
  struct cxiBufHeadChunk_t* bhcHeadP;

  /* Ignore if bi_size is still non-zero */
  if (bioP->bi_size)
    return 1;

  /* Decrement number of outstanding bios */
  bhcHeadP = (struct cxiBufHeadChunk_t*)bioP->bi_private;
  atomic_dec(&bhcHeadP->nIODoneLeft);

  /* If this is the last bio in the chunk list, enqueue the chunk list
     on the completion queue */
  if (atomic_read(&bhcHeadP->nIODoneLeft) == 0)
    chunkListDone(bhcHeadP);

  return 0;
}


/* Start a read or write of the given sectors from dev.  Data should be
   placed into the I/O buffer beginning at address userAddrP.  Returns 0
   on success, errno values on error. */
/* Linux 2.6 version */
int
cxiStartIO(int ioVecSize,
           cxiIOVecEntry* iovP,
           Boolean isWrite, cxiDev_t dev,
           UInt64 startSector, int sectorSize,
           void* qP,
           cxiIODoneIntCallback_t intCallbackRtn,
           void* callbackParmP)
{
  struct block_device* bdevP;
  int maxIOVec;
  int vecsAllocated;
  int nTotalPages;
  int iovIndex;
  int pageOffset;
  int sectorsThisPage;
  int i;
  int nSectors;
  int nBytes;
  unsigned long bufOffset;
  unsigned long bufEndOffset;
  struct bio* bioP;
  struct page* pageP;
  struct cxiBufHeadChunk_t* bhcP;
  struct cxiBufHeadChunk_t* bhcHeadP;
  struct cxiBufHeadChunk_t* bhcTailP;
  struct cxiIODoneQueue_t* queueP = (struct cxiIODoneQueue_t*)qP;
  request_queue_t* reqQP;
  cxiXmem_t* xmemDescP;
  char* userAddrP;

  ENTER(0);
  TRACE5(TRACE_IO, 5, TRCID_KDOIO_LINUX_BIO,
         "cxiStartIO: uaddr 0x%lX isWrite %d dev 0x%X sector %llu queueP 0x%lX\n",
         iovP[0].baseEffAddr, isWrite, dev, startSector, queueP);

  /* Validate parameters */
  bdevP = bdget(new_decode_dev(dev));
  LOGASSERT(bdevP != NULL && bdevP->bd_disk != NULL);
  maxIOVec = bio_get_nr_vecs(bdevP); /* query max device vectors */

  /* Allocate a single chunk and link it to itself.  Subsequent chunks
     may be needed and are allocated below.  */
  bhcP = (struct cxiBufHeadChunk_t*)kmem_cache_alloc(BhcCacheP, SLAB_KERNEL);
  bhcHeadP = bhcP;
  bhcTailP = bhcP;
  if (bhcP == NULL)
    goto enomem;
  INIT_LIST_HEAD(&bhcP->bhcDoneLink);
  bhcP->bhcCompletionQueueP = queueP;
  bhcP->bhcIntCallbackRtn = intCallbackRtn;
  bhcP->bhcCallbackDataP = callbackParmP;
  bhcP->bhcNextP = bhcP; /* circular link to itself */
  bhcP->nSlotsUsed = 0;

  /* Initialize number of I/Os remaining to 1 to prevent bioDone from
     deciding that the I/O is completely done before all of its chunks have
     been submitted. */
  atomic_set(&bhcP->nIODoneLeft, 1);

  /* Generate one or more bio structs for each buffer in the input ioVec */
#ifdef VDISK_TODO
This loop in cxiStartIO makes at least one bio for each input buffer.  This
might not be necessary if bios can hold more than an input buffer s worth
of pages.
#endif
  for (i=0; i<ioVecSize; i++)
  {
    xmemDescP = iovP[i].xmemDescP;
    userAddrP = iovP[i].baseEffAddr;
    nSectors = iovP[i].nSectors;
    DBGASSERT(nSectors > 0);
    nBytes = iovP[i].nBytes;

    /* Compute the total number of pages spanned by the portion of the
       buffer that will participate in the I/O.  This equals the number of
       Linux io vectors needed.  */
    bufOffset = userAddrP -
                (char*)(((struct cxiPageList_t*)(xmemDescP->pageListP))->plBaseAddr);
    bufEndOffset = bufOffset + nBytes - 1;
    nTotalPages = (bufEndOffset/PAGE_SIZE) - (bufOffset/PAGE_SIZE) + 1;
    pageOffset = bufOffset % PAGE_SIZE;
    DBGASSERT((pageOffset%512) == 0);

    TRACE7(TRACE_IO, 6, TRCID_KDOIO_LINUX_IOVEC,
           "cxiStartIO: i %d userAddrP 0x%lX nBytes %d bufOffset 0x%lX"
           "bufEndOffset 0x%lX nTotalPages %d pageOffset 0x%X\n",
            i, userAddrP, nBytes, bufOffset, bufEndOffset, nTotalPages,
            pageOffset);

    /* Build bios and submit them */
    while (nSectors > 0)
    {
      vecsAllocated = MIN(nTotalPages, maxIOVec);
      bioP = bio_alloc(GFP_NOIO, vecsAllocated);
      if (bioP == NULL)
        goto enomem;

      /* Get another chunk of bio pointers if necessary */
      if (bhcP->nSlotsUsed == BIOS_PER_CHUNK)
      {
        bhcP = (struct cxiBufHeadChunk_t*)kmem_cache_alloc(BhcCacheP,
                                                           SLAB_KERNEL);
        if (bhcP == NULL)
          goto enomem;
        bhcP->nSlotsUsed = 0;

        bhcP->bhcNextP = bhcHeadP;
        bhcTailP->bhcNextP = bhcP;
        bhcTailP = bhcP;
      }

      /* Use next available pointer slot and increment used count */
      bhcP->biop[bhcP->nSlotsUsed] = bioP;
      bhcP->nSlotsUsed += 1;

      /* Fill in bio */
      iovIndex = 0;
      bioP->bi_vcnt = 0;   /* accumulated below as number of bi_io_vecs */
      bioP->bi_idx = 0;    /* used by lower layer for recording current index */
      bioP->bi_size = 0;
      bioP->bi_bdev = bdevP;
      bioP->bi_end_io = bioDone;
      bioP->bi_sector = startSector;

      /* bio points to head of chunk list */
      bioP->bi_private = (void *)bhcHeadP;

      sectorsThisPage = MIN((PAGE_SIZE-pageOffset) / 512, nSectors);
      while (iovIndex < vecsAllocated)
      {
        cxiGetPagePtrs((struct cxiPageList_t*)(xmemDescP->pageListP),
                       (unsigned long)userAddrP & ~(PAGE_SIZE-1), 1, &pageP);
        DBGASSERT(pageP != NULL);

        bioP->bi_io_vec[iovIndex].bv_page = pageP;
        bioP->bi_io_vec[iovIndex].bv_len = sectorsThisPage * 512;
        bioP->bi_io_vec[iovIndex].bv_offset = pageOffset;
        TRACE9(TRACE_IO, 6, TRCID_KDOIO_LINUX_BIO_PAGE,
               "cxiStartIO: addr 0x%lX bhcP 0x%lX bioP 0x%lX index %d sector"
               "%llu sectorsThisPage %d pageP 0x%lX pageOffset 0x%X maxIOVec %d\n",
               userAddrP, bhcP, bioP, iovIndex, startSector, sectorsThisPage,
               pageP, pageOffset, maxIOVec);
        userAddrP += (sectorsThisPage * 512);
        iovIndex += 1;

        bioP->bi_vcnt = iovIndex;
        bioP->bi_size += (sectorsThisPage * 512);

        /* Advance to next page */
        startSector += sectorsThisPage;
        nSectors -= sectorsThisPage;
        sectorsThisPage = MIN(nSectors, PAGE_SIZE/512);
        pageOffset = 0;
      }

      nTotalPages -= bioP->bi_vcnt;

      atomic_inc(&bhcHeadP->nIODoneLeft);
      submit_bio(isWrite, bioP);
    }  /* end of 'for all bios for the current iovec entry' */
  }  /* end of 'for all iovec entries' */

  /* Unplug the device queue to avoid 3ms delay when no other I/O in
     progress on the device */
  reqQP = bdev_get_queue(bdevP);
  if (reqQP->unplug_fn != NULL)
    reqQP->unplug_fn(reqQP);

  /* Remove the extra count put in bhcHeadP->nIODoneLeft at the
     beginning of this routine.  If the count is now zero, perform chunk
     done handling */
  atomic_dec(&bhcHeadP->nIODoneLeft);
  /* If the last I/O completion interrupt occurs right here, then
     chunkListDone will be called twice */
  if (atomic_read(&bhcHeadP->nIODoneLeft) == 0)
    chunkListDone(bhcHeadP);

  EXIT(0);
  return 0;

  /* Failed to allocate memory.  Difficult to clean up, since some I/Os
     may have already been started. */
enomem:
  LOGASSERT(!"Cannot allocate memory in cxiStartIO");
  return ENOMEM;
}


/* Routine to set up the disk block size and get disk parameters */
int 
GetDiskInfoX(cxiDev_t devId, struct cxiDiskInfo_t* diskInfoP)
{
  struct block_device *bdevP = bdget(new_decode_dev(devId));

  ENTER(0);
  LOGASSERT(bdevP != NULL && bdevP->bd_disk != NULL);

  diskInfoP->sectorSize = bdev_hardsect_size(bdevP);

  if (diskInfoP->sectorSize == 0)
    diskInfoP->sectorSize = 512;

  /* Set blocksize of this device to hardware sector size */
  set_blocksize(bdevP, diskInfoP->sectorSize);

  DBGASSERT(bdevP->bd_inode != NULL);
  diskInfoP->totalSectors = bdevP->bd_inode->i_size / diskInfoP->sectorSize;

  TRACE3(TRACE_IO, 2, TRCID_GET_DISKINFOX,
         "GetDiskInfoX: devId %08llX sector size %d totalSectors %lld\n",
         devId, diskInfoP->sectorSize, diskInfoP->totalSectors);
  EXIT(0);
  return 0;
}
#endif


/* Wait for the given number of asynchronous I/Os to complete.  The
   callback routine will be invoked for each completed I/O in the
   context of the calling thread to set the error code in higher level
   data structures.  Free the chunk lists and their associated bios or
   buffer_heads. */
void cxiWaitIO(struct cxiIODoneQueue_t* queueP, int nDesired,
               cxiIODoneThreadCallback_t threadCallbackRtn)
{
  unsigned long flags;
  struct list_head * p;
  struct cxiBufHeadChunk_t*  bhcHeadP;
  struct cxiBufHeadChunk_t*  bhcP;
  struct cxiBufHeadChunk_t*  nextBhcP;
  void* callbackParmP;
  int i;
  int upToDate;

  ENTER(0);
  TRACE3(TRACE_IO, 6, TRCID_WAITIO_ENTER,
         "cxiWaitIO: queueP 0x%lX nDesired %d ioqNQueued %d\n",
         queueP, nDesired, atomic_read(&queueP->ioqNQueued));

  /* Record the number of desired I/O completions */
  atomic_set(&queueP->ioqNDesired, nDesired);

  /* Wait until enough I/Os have completed */
  wait_event(queueP->ioqWaitQ, \
             (atomic_read(&queueP->ioqNQueued) >= atomic_read(&queueP->ioqNDesired)));

  /* With interrupts disabled, save the current queue contents and reset it
     to empty.  Note that the first and last queued entries still point at
     the original queue header, so normal list manipulation routines cannot
     be used on the saved queue. */
  SPIN_LOCK_IRQ(&queueP->ioqWaitQ.lock, flags);
  p = queueP->ioqListHead.next;
  INIT_LIST_HEAD(&queueP->ioqListHead);
  atomic_set(&queueP->ioqNQueued, 0);
  atomic_set(&queueP->ioqNDesired, MAX_IO_CHUNK_LISTS);
  SPIN_UNLOCK_IRQ(&queueP->ioqWaitQ.lock, flags);

  /* Deallocate each chunk list whose I/O has completed */
  while (p != &queueP->ioqListHead)
  {
    /* Convert the list_head pointer into a pointer to the head of a
       list of chunks, and update the list_head pointer so the chunk
       list head can be deallocated */
    bhcHeadP = list_entry(p, struct cxiBufHeadChunk_t, bhcDoneLink);
    p = p->next;
    LOGASSERT(atomic_read(&bhcHeadP->nIODoneLeft) == 0);
    DBGASSERT(bhcHeadP->bhcCompletionQueueP == queueP);

    /* Walk the list of chunks, accumulating the return code and freeing
       each chunk */
    callbackParmP = bhcHeadP->bhcCallbackDataP;
    bhcP = bhcHeadP;
    upToDate = 1;
    do
    {
      TRACE2(TRACE_IO, 6, TRCID_WAITIO_CHUNK,
             "cxiWaitIO: freeing bhcP 0x%lX nSlotsUsed %d\n",
             bhcP, bhcP->nSlotsUsed);

      /* Scan all bios or buffer_heads in this chunk and compute whether
         or not any I/O errors were encountered */
      for (i=0; i<bhcP->nSlotsUsed; i++)
      {
#       if LINUX_KERNEL_VERSION < 2050000
          if (!buffer_uptodate(&bhcP->bh[i]))
            upToDate = 0;
#       else
          if (!test_bit(BIO_UPTODATE, &bhcP->biop[i]->bi_flags))
            upToDate = 0;
          bio_put(bhcP->biop[i]);
          bhcP->biop[i] = NULL;
#       endif
      }

      /* Remember the next chunk on the list, then deallocate the current
         chunk */
      nextBhcP = bhcP->bhcNextP;
      kmem_cache_free(BhcCacheP, (void*)bhcP);

      /* Advance to next chunk */
      bhcP = nextBhcP;
    } while (bhcP != bhcHeadP);

    /* Set return code in higher-level data structures */
    threadCallbackRtn(callbackParmP, upToDate==1 ? 0 : EIO);
  }
  EXIT(0);
}


/* Initialize an I/O completion queue */
void cxiInitIODoneQueue(struct cxiIODoneQueue_t* queueP)
{
  init_waitqueue_head(&queueP->ioqWaitQ);
  atomic_set(&queueP->ioqNQueued, 0);
  atomic_set(&queueP->ioqNDesired, MAX_IO_CHUNK_LISTS);
  INIT_LIST_HEAD(&queueP->ioqListHead);
}


/* Terminate an I/O completion queue */
void cxiTermIODoneQueue(struct cxiIODoneQueue_t* queueP)
{
}


/* Allocate an I/O completion queue */
struct cxiIODoneQueue_t* cxiAllocIOWaitQueue()
{
  struct cxiIODoneQueue_t* queueP;

  queueP = (struct cxiIODoneQueue_t*)
             cxiMallocPinned(sizeof(struct cxiIODoneQueue_t));

  if (queueP != NULL)
  {
    cxiInitIODoneQueue(queueP);
  }
  return queueP;
}


/* Free an I/O completion queue */
void cxiFreeIOWaitQueue(struct cxiIODoneQueue_t* queueP)
{
  DBGASSERT(atomic_read(&queueP->ioqNQueued) == 0);
  cxiFreePinned((void*)queueP);
}


/* Allocate and initialize an array of I/O completion queues */
int cxiAllocIOCompletionQueueArray(int nQueues, char** basePP, int* queueSizeP)
{
  char* baseP;
  struct cxiIODoneQueue_t* queueP;
  int i;
  int err=0;

  *queueSizeP = sizeof(struct cxiIODoneQueue_t);
  baseP = (char*)cxiMallocPinned(nQueues * sizeof(struct cxiIODoneQueue_t));
  if (baseP == NULL)
  {
    err = ENOMEM;
    goto exit;
  }
  queueP = (struct cxiIODoneQueue_t*)baseP;
  for (i=0; i<nQueues; i++)
  {
    cxiInitIODoneQueue(queueP);
    queueP += 1;
  }
  *basePP = baseP;

exit:
  TRACE1(TRACE_IO, 3, TRCID_ALLOC_IOCQ_ARRAY_LINUX,
         "cxiAllocIOCompletionQueueArray: baseP 0x%lX\n", baseP);
  return err;
}


/* Free an array of I/O completion queues that were allocated by
   cxiAllocIOCompletionQueueArray */
void cxiFreeIOCompletionQueueArray(int nQueues, char* freeP)
{
  cxiFreePinned((void*)freeP);
}

